﻿using UnityEngine;
using UnityEngine.SceneManagement;
using System.Collections;

// By Sergey Vitukhin. 20150205.

public class PiScene : MonoBehaviour
{
    private enum LoadState : byte
    {
        None = 0,       //Ready for Load
        FadeIn = 1,     //FadeIn before load
        PreClean = 2,   //GC
        Load = 3,       //Load Level
        PostWait = 4,   //Initing Level in other scripts
        PostLoad = 5,   //Resources unload
        PostClean = 6,  //GC
        Black = 7,      //Loading
        FadeOut = 8,    //FadeOut after load
        White = 9       //Loading complete
    }

    public static System.Action<int> OnLoadBegin;
    public static System.Action OnLoadEnd;

    private static PiScene _instance;
    private static readonly string _name = "PiScene";
    private static bool _isInitialized = false;

    [Range(0f, 3f)]
    public float fadeInTime = 1f;
    [Range(0, 60)]
    public int skipFrames = 10;
    [Range(0f, 3f)]
    public float blackTime = 0.5f;
    [Range(0f, 3f)]
    public float quickBlackTime = 0.5f;
    [Range(0f, 3f)]
    public float fadeOutTime = 1f;
    [Range(0f, 3f)]
    public float whiteTime = 0f;

    private LoadState _state = LoadState.None;
    protected int _sceneID = -1;
    private float _currentTime = 0f;
    private int _currentSkipFrames = 0;
    private float _blackTime = 0f;
    private AsyncOperation _resourceUnloadTask = null;

    #region Static Methods
    public static void Initialize()
    {
        if (_isInitialized)
            return;

        Instantiate(Resources.Load(_name));
    }

    public static void Load(int sceneID, bool quick)
    {
        if (!_isInitialized)
            Initialize();

        if (_isInitialized && _instance != null && _instance._state == LoadState.None)
        {
            _instance._sceneID = sceneID;
            _instance._state = LoadState.FadeIn;
            _instance._currentTime = 0f;
            _instance._currentSkipFrames = _instance.skipFrames;
            _instance._blackTime = (quick) ? _instance.quickBlackTime : _instance.blackTime;
        }
    }
    #endregion

    #region ToOverride
    protected virtual void FadeIn(float value)
    {
        if (value == 1f)
        {
            _state = LoadState.PreClean;
            _currentTime = 0f;
            if (OnLoadBegin != null)
                OnLoadBegin(_sceneID);
        }
    }

    private void PreClean()
    {
        System.GC.Collect();
        _state = LoadState.Load;
    }
    
    private void Load()
    {
        SceneManager.LoadScene(_sceneID, LoadSceneMode.Single);
        _state = LoadState.PostWait;
    }

    private void PostWait()
    {
        _state = LoadState.PostLoad;
    }

    private void PostLoad()
    {
        if (_resourceUnloadTask == null)
        {
            _resourceUnloadTask = Resources.UnloadUnusedAssets();
        }
        else
        {
            if (_resourceUnloadTask.isDone == true)
            {
                _resourceUnloadTask = null;
                _state = LoadState.PostClean;
            }
        }
    }
    
    private void PostClean()
    {
        System.GC.Collect();
        _state = LoadState.Black;
    }

    protected virtual void Black(float value)
    {
        if (value == 1f)
        {
            _state = LoadState.FadeOut;
            _currentTime = 0f;
        }
    }

    protected virtual void FadeOut(float value)
    {
        if (value == 1f)
        {
            _state = LoadState.White;
            _currentTime = 0f;
        }
    }

    protected virtual void White(float value)
    {
        if (value == 1f)
        {
            _state = LoadState.None;
            if (OnLoadEnd != null)
                OnLoadEnd();
        }
    }
    #endregion

    #region MonoBehaviour
    protected void Awake()
    {
        if (!_isInitialized && _instance == null)
        {
            _instance = this;
            DontDestroyOnLoad(gameObject);
            name = _name;
            _isInitialized = true;
        }
        else
        {
            Destroy(gameObject);
        }
    }

    protected void OnDestroy()
    {
        if (_instance == this)
        {
            _instance = null;
            OnLoadBegin = null;
            OnLoadEnd = null;
        }
    }

    protected void Update()
    {
        switch (_state)
        {
            case LoadState.FadeIn:
                _currentTime += Time.unscaledDeltaTime;
                FadeIn((_currentTime < fadeInTime && fadeInTime > 0f) ? (_currentTime / fadeInTime) : 1f);
                break;
            
            case LoadState.PreClean:
                PreClean();
                break;
            
            case LoadState.Load:
                Load();
                break;
            
            case LoadState.PostWait:
                //Need to whait in frames - deltaTime can be out of a value
                if (_currentSkipFrames > 0)
                {
                    --_currentSkipFrames;
                    return;
                }
                PostWait();
                break;
            
            case LoadState.PostLoad:
                PostLoad();
                break;
            
            case LoadState.PostClean:
                PostClean();
                break;
            
            case LoadState.Black:
                _currentTime += Time.unscaledDeltaTime;
                Black((_currentTime < _blackTime && _blackTime > 0f) ? (_currentTime / _blackTime) : 1f);
                break;
            
            case LoadState.FadeOut:
                _currentTime += Time.unscaledDeltaTime;
                FadeOut((_currentTime < fadeOutTime && fadeOutTime > 0f) ? (_currentTime / fadeOutTime) : 1f);
                break;
            
            case LoadState.White:
                _currentTime += Time.unscaledDeltaTime;
                White((_currentTime < whiteTime && whiteTime > 0f) ? (_currentTime / whiteTime) : 1f);
                break;
            
            default:
                break;
        }
    }
    #endregion
}